;;; "2-fix" syntax reader macro, with associated functions
;;; by Dan Bensen
#|
-------------------------------------------------------------------------
This software is Copyright (c) 2008 Daniel S. Bensen.
Permission is granted to use, copy, modify, and distribute this software,
provided that this copyright and permission notice is included in full
in all copies and supporting documentation.
This software is provided "as is" with no express or implied warranty.
-------------------------------------------------------------------------
|#

;;(defpackage :2fix (:use :cl)
;;	    (:export isatom islist iscons ischar isnum isint
;;		     += *= or= and= not= xor=
;;		     ^ ^2 ^3 ^4 ^5 ^6 ^7 ^8 def-op= =0 >0 <0 >=0 <=0 !=0))
;;(in-package :2fix)

(in-package :std)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defun |read-[2-fix]| (stream char) 
    (declare (ignore char)) 
    (let* ((list (read-delimited-list #\] stream t))
	   (op (cadr list)))
      ;;(unless op (error "No operator in 2-fix: ~S" list))
      (list* op (car list) (cddr list))))
  (set-macro-character #\[ #'|read-[2-fix]|         nil) 
  (set-macro-character #\] (get-macro-character #\) nil)))

(macrolet
    ((defx-synonyms (&body pairs)
	 (let ((x (gensym)))
	   `(progn
	     ,@(loop for pair in pairs
		     collect `(progn 
			       (defun ,(car pair) (,x)
				 (declare (inline ,(cadr pair)))
				 (,(cadr pair) ,x))
			       (export ',(car pair))))))))
  (defx-synonyms
      (is-adjustable-array adjustable-array-p)
      (is-alpha-char alpha-char-p)
    (isalphanum alphanumericp)
    (has-array-fill-pointer array-has-fill-pointer-p)
    (is-array-in-bounds array-in-bounds-p)
    (isatom atom)
    (is-bit-vector bit-vector-p)
    (is-both-case both-case-p)
    (isbound boundp)
    (ischar characterp)
    ;;  (iscommon commonp)
    (is-compiled-func compiled-function-p)
    (iscomplex complexp)
    (iscons consp)
    (isconstant constantp)
    (isdigchar digit-char-p)
    (isend endp)
    (iseven evenp)
    (isfbound fboundp)
    (isfloat floatp)
    (isfunc functionp)
    (is-hashtable hash-table-p)
    (iskeyword keywordp)
    (is-lower-case lower-case-p)
    (isnum  numberp)
    (isodd oddp)
    (ispackage packagep)
    (ispathname pathnamep)
    (isrational rationalp)
    (isreadtable readtablep)
    (isreal realp)
    (is-simple-string simple-string-p)
    (is-simple-vector simple-vector-p)
    ;; (slot-exists slot-exists-p)
    ;; (is-special-form special-form-p)
    (isstream streamp)
    (isstring stringp)
    (is-symbol symbolp)
    (is-upper-case upper-case-p)
    (isvector vectorp)
    (is-wild-pathname wild-pathname-p)
    (=0 zerop)
    (>0 plusp)
    (<0 minusp)
    ))

(macrolet ((defx-synonyms-2 (&body pairs)
  (let ((x (gensym))
	(y (gensym)))
    `(progn
      ,@(loop for pair in pairs
	      collect `(progn
			(defun ,(car pair) (,x ,y)
			(declare (inline ,(cadr pair)))
			(,(cadr pair) ,x ,y))
			(export ',(car pair))))))))
(defx-synonyms-2
  (isequal equalp)
  (istail tailp)
  (issubtype subtypep)
  (has-type typep)
))

(macrolet ((defx-synonyms-2+ (&body pairs)
  (let ((x  (gensym))
	(y  (gensym))
	(ys (gensym)))
    `(progn
      ,@(loop for pair in pairs
	      collect `(progn
			(defun ,(car pair) (,x ,y &rest ,ys)
			  (declare (inline ,(cadr pair)))
			  (apply #',(cadr pair) ,x ,y ,ys))
			(export ',(car pair))))))))
  (defx-synonyms-2+
      (is-char-greater   char-greaterp  ) (not-char-greater  char-not-greaterp   )
    (is-char-less      char-lessp     ) (not-char-less      char-not-lessp     )
    (is-string-greater string-greaterp) (not-string-greater string-not-greaterp)
    (is-string-less    string-lessp   ) (not-string-less    string-not-lessp   )
    (issubset subsetp)
    ))
  
(macrolet ((defx (f args &body body)
	       `(progn
		 (defun ,f ,args ,@body)
		 (export ',f)))
	   (def-op= (op= op)
	       `(progn
		 (defmacro ,op= (place val)
		   `(setf ,place [,place ,',op ,val]))
		 (export ',op=))))
  (defx !=0 (x) (not (zerop  x)))
  (defx <=0 (x) (not (plusp  x)))
  (defx >=0 (x) (not (minusp x)))
  
  (def-op=  += +) (def-op=  -= -)
  (def-op=  *= *) (def-op= //= /)
  
  (def-op=  or= orf)
  (def-op= and= andf)
  (def-op= not= not)
  (def-op= xor= xorf)
  
  (defx ^ (a b) (expt a b))
  (defx ^2 (x) (* x x))
  (defx ^3 (x) (* x x x))
  (defx ^4 (x) (let ((x^2 [x * x])) [x^2 * x^2]))
  (defx ^5 (x) (let ((x^2 [x * x])) [x * [x^2 * x^2]]))
  
  (defx ^6 (x) (let* ((x^2 [x * x]) (x^4 [x^2 * x^2]))  [x^4 * x^2]))
  (defx ^7 (x) (let* ((x^2 [x * x]) (x^4 [x^2 * x^2])) [[x^4 * x^2] * x]))
  (defx ^8 (x) (let* ((x^2 [x * x]) (x^4 [x^2 * x^2]))  [x^4 * x^4]))
  )
